//
// Informa -- RSS Library for Java
// Copyright (c) 2002 by Niko Schmuck
//
// Niko Schmuck
// http://sourceforge.net/projects/informa
// mailto:niko_schmuck@users.sourceforge.net
//
// This library is free software.
//
// You may redistribute it and/or modify it under the terms of the GNU
// Lesser General Public License as published by the Free Software Foundation.
//
// Version 2.1 of the license should be included with this distribution in
// the file LICENSE. If the license is not included with this distribution,
// you may find a copy at the FSF web site at 'www.gnu.org' or 'www.fsf.org',
// or you may write to the Free Software Foundation, 675 Mass Ave, Cambridge,
// MA 02139 USA.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied waranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//

// $Id: ChannelBuilder.java,v 1.25 2004/07/10 09:11:06 spyromus Exp $

package de.nava.informa.impl.hibernate;

import java.net.URL;
import java.util.*;

import net.sf.hibernate.*;

import org.apache.commons.logging.*;
import org.jdom.Element;

import de.nava.informa.core.*;
import de.nava.informa.utils.InformaUtils;

/**
 * Factory for the creation of the channel object model with the hibernate
 * persistent store.
 *
 * NOT THREAD SAFE
 *
 * Hibernate Multi-threading notes: ChannelBuilder has some subtleties as it
 * relates to threading. The specifics of the way it is supported
 * still need to be proven. Certainly the error handling here and in
 * UpdateChannelTask and in ChannelRegistry is incomplete. It seems to work,
 * but I would consider it incomplete still.
 *
 * The key facts are (1) Sessions are not thread safe and (2) Sessions should
 * have relatively short lifespans.
 *
 * To support this, there is a mode of using ChannelBuilder where it holds on
 * to a SessionHandler and manages the creation and destruction of Sessions on
 * behalf of the caller. When you supply a SessionHandler to ChannelBuilder,
 * you may use the beginTransaction() and endTransaction() calls to take all the
 * steps needed before and after a transaction. At the end of endTransaction()
 * the transaction will be closed and the session will be flushed and closed.
 * To use this mode, you should (1) Create a SessionHandler , (2) Create a
 * JDBC Connection to the database, (3) sessionHandler.setConnection(connection),
 * and (4) use new ChannelBuilder(sessionHandler).
 *
 * @author Niko Schmuck (niko@nava.de)
 */
public class ChannelBuilder implements ChannelBuilderIF {

  private static Log logger = LogFactory.getLog(ChannelBuilder.class);
  private Session session;
  private SessionHandler handler;
  private Transaction transaction;

  /**
   * ChannelBuilder constructor. Caller is responsible for managing sessions
   * and transactions.
   */
  public ChannelBuilder(Session session) {
    logger.info("New Channel Builder for: " + session);
    this.session = session;
    this.handler = null;
  }

  /**
   * ChannelBuilder constructor. ChannelBuilder will manage sessions and
   * transactions. Supplied SessionHandler needs to have a live JDBC connection
   * available.
   */
  public ChannelBuilder(SessionHandler handler) {
    logger.debug("New Channel Builder for: " + handler);
    this.handler = handler;
    this.session = null;
  }

  // --------------------------------------------------------------
  // Hibernate Specific Methods
  // --------------------------------------------------------------

  /**
   * Processing needed at the start of a transaction.
   *  - creating a session
   *  - beginning the transaction
   */
  public void beginTransaction() throws ChannelBuilderException {
    logger.info("beginTransaction");
    if (session != null || handler == null)
      throw new IllegalStateException("Session != null || handler == null");
    try {
      session = handler.getSession();
      transaction = session.beginTransaction();
    } catch (HibernateException e) {
      e.printStackTrace();
      transaction = null;
      throw new ChannelBuilderException(e);
    }
  }

  /**
   * Processing needed at the end of a transaction.
   *   - commit the transaction
   *   - flush the session
   *   - close the session
   * TODO: catch the exception so this method doesn't have any throws.
   */
  public void endTransaction() throws ChannelBuilderException {
    logger.info("endTransaction");
    if (handler == null || transaction == null || session == null)
      throw new IllegalStateException("handler == null || transaction == null || session == null");
    try {
      transaction.commit();
      session.flush();
      session.close();
      session = null;
      transaction = null;

    } catch (HibernateException he) {
      if (transaction != null)
        try {
          he.printStackTrace();
          transaction.rollback();
          transaction = null;
          if (session.isOpen()) {
            session.close();
            session = null;
          }
        } catch (HibernateException e) {
          if (session.isOpen()) {
            session = null;
          }
          e.printStackTrace();
          throw new ChannelBuilderException(e);
        }
      throw new ChannelBuilderException(he);
    }
  }

  /**
   * Check if we are already in the middle of a transaction. This is needed because
   * as of now begin/endTransactions cannot be nested and in fact give assert errors
   * if you try.
   *
   * @return - boolean indicating whether we are currently in a transaction.
   */
  public boolean inTransaction()
  {
    return session != null && transaction != null;
  }
  /**
   * resetTransaction - Used during error handling. If in a catch block there
   * is a potentially still open transaction (i.e. beginTransaction() was called)
   * then call this method to reset the state of the ChannelBuilder and clean up
   * the transaction.
   *
   */
  public void resetTransaction()
  {
    logger.debug("Transaction being reset.");
    if (transaction != null)
    {
      try {
        transaction.commit();
        transaction = null;
      }
      catch (HibernateException e) {
        transaction = null;
        e.printStackTrace();
      }
    }
    if (session != null) {
      try {
        session.flush();
        session.close();
        session = null;
     }
      catch (HibernateException e) {
        e.printStackTrace();
        session = null;
      }
    }
  }

  /**
   * Certain Hibernate calls require the session. Note that this
   * call should only be made between a beginTransaction and endTransaction call
   * which is why we throw an IllegalStateException otherwise.
   */
  public Session getSession() {
    if (handler == null || session == null)
      throw new IllegalStateException("getSession must be bracketed by begin/endTransaction");
    if (!handler.isSessionOpen())
        throw new IllegalStateException("Hibernate Handler must be open");
    return session;
  }

  /**
   * update - Hibernate Update some object
   *
   * @param o
   * @throws ChannelBuilderException -
   */
  public void update(Object o) throws ChannelBuilderException {
    try {
      session.update(o);
    } catch (HibernateException e) {
      e.printStackTrace();
      throw new ChannelBuilderException("update() Failed");
    }
  }

  /**
   * uHibernate Delete some object
   *
   * @param o - Object to Delete
   * @throws ChannelBuilderException - Translation of Hibernate exception
   */
  public void delete(Object o) throws ChannelBuilderException {
    try {
      session.delete(o);
    } catch (HibernateException e) {
      e.printStackTrace();
      throw new ChannelBuilderException("delete() Failed");
    }
  }


  // --------------------------------------------------------------
  // implementation of ChannelBuilderIF interface
  // --------------------------------------------------------------

  public void init(Properties props) throws ChannelBuilderException {
    logger.debug("initialising channel builder for hibernate backend");
  }

  public ChannelGroupIF createChannelGroup(String title) {
    ChannelGroupIF obj = new ChannelGroup(title);
    save(obj);
    return obj;
  }

  public ChannelIF createChannel(String title) {
    ChannelIF obj = new Channel(title);
    save(obj);
    return obj;
  }

  public ChannelIF createChannel(Element channelElement, String title) {
    ChannelIF obj = new Channel(channelElement, title);
    save(obj);
    return obj;
  }

  public ItemIF createItem(ChannelIF channel, String title, String description,
                           URL link) {
    return createItem(null, channel, title, description, link);
  }

  public ItemIF createItem(Element itemElement, ChannelIF channel,
                           String title, String description, URL link) {
    ItemIF obj = new Item(itemElement, channel, title, description, link);
    save(obj);
    if (channel != null) {
      channel.addItem(obj);
    }
    return obj;
  }

  /**
   * Create an item from an existing item.
   *
   * @TODO: Add code to copy the complete state of the item rather than just four fields!
   *
   * @param chan - Channel in which the new item will be held
   * @param oldItem - Old item which will provide the state for the new item
   * @return - the newly created item
   */
  public ItemIF createItem(ChannelIF chan, ItemIF oldItem) {
    ItemIF obj = createItem(null, chan, oldItem.getTitle(), oldItem.getDescription(), oldItem.getLink());
    InformaUtils.copyItemProperties(oldItem, obj);
    save(obj);
    return obj;
  }

  public ImageIF createImage(String title, URL location, URL link) {
    ImageIF obj = new Image(title, location, link);
    save(obj);
    return obj;
  }

  public CloudIF createCloud(String domain, int port, String path, String registerProcedure, String protocol) {
    logger.info("ChannelBuilder is creating a Persistent Cloud");
    CloudIF obj = new Cloud(domain, port, path, registerProcedure, protocol);
    save(obj);
    return obj;
  }

  public TextInputIF createTextInput(String title, String description,
                                     String name, URL link) {
    TextInputIF obj = new TextInput(title, description, name, link);
    save(obj);
    return obj;
  }

  public ItemSourceIF createItemSource(ItemIF item, String name,
                                       String location, Date timestamp) {
    return new ItemSource(item, name, location, timestamp);
  }

  public ItemEnclosureIF createItemEnclosure(ItemIF item, URL location,
                                             String type, int length) {
    return new ItemEnclosure(item, location, type, length);
  }

  public ItemGuidIF createItemGuid(ItemIF item, String location, boolean permaLink) {
    return new ItemGuid(item, location, permaLink);
  }

  public CategoryIF createCategory(CategoryIF parent, String title) {
    CategoryIF cat = new Category(title);
    save(cat);
    if (parent != null) {
      parent.addChild(cat);
    }
    return cat;
  }

  public void close() throws ChannelBuilderException {
    logger.debug("closing channel builder for hibernate backend");
  }

  /**
   * Reloads group for use in new session.
   *
   * @param group to reload.
   *
   * @return reloaded group for chaning.
   *
   * @throws ChannelBuilderException when unable to reload data.
   */
  public ChannelGroup reload(ChannelGroup group)
    throws ChannelBuilderException
  {
    try
    {
      getSession().load(group, new Integer(group.getIntId()));
    } catch (HibernateException e)
    {
      throw new ChannelBuilderException("Unable to reload group: " + e.getMessage());
    }

    return group;
  }

  /**
   * Reloads channel for use in new session.
   *
   * @param channel channel to reload.
   *
   * @return reloaded channel for chaining.
   *
   * @throws ChannelBuilderException when unable to reload data.
   */
  public Channel reload(Channel channel)
    throws ChannelBuilderException
  {
    try
    {
      getSession().load(channel, new Integer(channel.getIntId()));
    } catch (HibernateException e)
    {
      throw new ChannelBuilderException("Unable to reload channel: " + e.getMessage());
    }

    return channel;
  }

  /**
   * Reloads item for use in new session.
   *
   * @param item  item to reload.
   *
   * @return reloaded item for chaning.
   *
   * @throws ChannelBuilderException when unable to reload data.
   */
  public Item reload(Item item)
    throws ChannelBuilderException
  {
    try
    {
      getSession().load(item, new Integer(item.getIntId()));
    } catch (HibernateException e)
    {
      throw new ChannelBuilderException("Unable to reload item: " + e.getMessage());
    }

    return item;
  }

  // -------------------------------------------------------------
  // internal helper methods
  // -------------------------------------------------------------

  protected void save(Object dataObject) {
    if (session == null)
      throw new IllegalStateException("Session == null");
    try {
      session.save(dataObject);
    } catch (HibernateException he) {
      throw new RuntimeException(he.getMessage());
    }
  }
}
